﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Net;
using System.Net.Sockets;
using System.Timers;

namespace OZX
{
    public class OzxClient
    {
        //*********************************************************
        // Public variables
        //*********************************************************

        public string Host;
        public int Port;
        public string Username;
        public string Password;
        public bool AutoReconnect;

        //*********************************************************
        // Private class variables
        //*********************************************************

        TcpClient client;
        byte[] _inputBuffer;
        object _sendLock;
        object _connectLock;
        DateTime _lastDataReceived;
        Timer _keepAliveTimer;
        Timer _reconnectTimer;
        bool _connectionInProgress;

        //*********************************************************
        // Construction
        //*********************************************************

        public OzxClient()
        {
            _inputBuffer = new byte[0];
            _connectLock = new object();
            _sendLock = new object();
            AutoReconnect = true;
        }

        //*********************************************************
        // Public interface
        //*********************************************************

        public event EventHandler<OzxArgs<IOzxMessage>> OnMessageReceived;
        public event EventHandler<OzxArgs<string>> OnMessageAcceptedForDelivery;
        public event EventHandler<OzxArgs<string, string>> OnMessageNotAcceptedForDelivery;
        public event EventHandler<OzxArgs<string, DateTime>> OnMessageSubmitSuccess;
        public event EventHandler<OzxArgs<string, DateTime, string>> OnMessageSubmitFailed;
        public event EventHandler<OzxArgs<string, DateTime>> OnMessageDeliverySuccess;
        public event EventHandler<OzxArgs<string, DateTime, string>> OnMessageDeliveryFailed;
        public event EventHandler<OzxArgs<string, DateTime>> OnMessageViewed;

        public event EventHandler<EventArgs> OnConnected;
        public event EventHandler<OzxArgs<string>> OnDisconnected;

        //*********************************************************
        // Networking / connect
        //*********************************************************

        public void Connect(string host, int port, string username, string password)
        {
            Host = host;
            Port = port;
            Username = username;
            Password = password;
            Connect();
        }

        public void Connect()
        {
            if (_connectionInProgress) return;
            lock (_connectLock)
            {
                if (_connectionInProgress) return;
                _connectionInProgress = true;
                stopReconnectTimer();
                if (client != null)
                    disconnect("Connection requested on an open connection.");
                try
                {
                    client = new TcpClient(AddressFamily.InterNetwork);
                    client.BeginConnect(Host, Port, new AsyncCallback(connectCallback), this);
                }
                catch (Exception e)
                {
                    disconnect(e.Message);
                }
            }
        }

        static void connectCallback(IAsyncResult ar)
        {
            OzxClient c = (OzxClient)ar.AsyncState;
            c._connectionInProgress = false;
            TcpClient t = c.client;
            try
            {
                t.EndConnect(ar);
            }
            catch (Exception e)
            {
                c.disconnect(e.Message);
                return;
            }
            c.beginRead();
            c.sendlogin();
            c.startKeepAlvie();
        }

        //*********************************************************
        // Networking / Disconnect
        //*********************************************************

        public void Disconnect()
        {
            disconnect("Graceful disconnect.");
            stopReconnectTimer();
        }

        void disconnect(string error)
        {
            lock (_connectLock)
            {
                if (client == null)
                    return;
                disposeClient();
            }
            OnDisconnected(this, new OzxArgs<string>(error));
            startReconnectTimer();
        }

        void disposeClient()
        {
            if (client == null) return;
            try
            {
                stopKeepAlive();
                client.Close();
                client = null;
            }
            catch
            {
            }
        }

        //*********************************************************
        // Networking / Receive data
        //*********************************************************

        void beginRead()
        {
            try
            {
                var buffer = new byte[4096];
                var ns = client.GetStream();
                ns.BeginRead(buffer, 0, buffer.Length, endRead, buffer);
            }
            catch (Exception e)
            {
                disconnect("Receive failed.");
            }
        }

        void endRead(IAsyncResult result)
        {
            try
            {
                var buffer = (byte[])result.AsyncState;
                if (client == null) return;
                var ns = client.GetStream();
                var bytesAvailable = ns.EndRead(result);

                if (bytesAvailable == 0)
                {
                    disconnect("Socket connection lost.");
                    return;
                }
                _lastDataReceived = DateTime.Now;
                var newInputBufferLength = _inputBuffer.Length + bytesAvailable;
                var newInputBuffer = new byte[newInputBufferLength];
                if (_inputBuffer.Length > 0)
                    Buffer.BlockCopy(_inputBuffer, 0, newInputBuffer, 0, _inputBuffer.Length);
                Buffer.BlockCopy(buffer, 0, newInputBuffer, _inputBuffer.Length, bytesAvailable);
                _inputBuffer = newInputBuffer;

                processInputBuffer();

                beginRead();
            }
            catch (Exception e)
            {
                string error = "Connection lost.";
                if (e.InnerException != null)
                    error = e.InnerException.Message;
                disconnect(error);
            }
        }

        void processInputBuffer()
        {
            if (_inputBuffer.Length < 1)
            {
                //not enough bytes
                return;
            }

            //Read pdu
            byte command = _inputBuffer[0];

            int step;
            byte[] newInputBuffer;
            if (_inputBuffer.Length < 5)
            {
                //not enough bytes
                return;
            }

            byte[] lenBytes = new byte[] { _inputBuffer[1], _inputBuffer[2], _inputBuffer[3], _inputBuffer[4] };
            int offset = 0;
            int len = OzCommonBytes.BytesToInt(lenBytes, ref offset);
            if (_inputBuffer.Length < 5 + len)
            {
                //not enough bytes
                return;
            }
            byte[] dataBytes = new byte[len];
            Buffer.BlockCopy(_inputBuffer, 5, dataBytes, 0, len);
            offset = 0;
            step = len + 5;

            //Decode data
            var cmd = OzPDU_OZX.GetCommandInstance(command);
            if (cmd != null)
            {
                OzPDU_OZX.DecodeData(dataBytes, ref offset, cmd);
                onPduReceived(cmd);
            }

            //Clear processed data from input buffer
            if (_inputBuffer.Length == step)
            {
                _inputBuffer = new byte[0];
                return;
            }
            newInputBuffer = new byte[_inputBuffer.Length - step];
            Buffer.BlockCopy(_inputBuffer, step, newInputBuffer, 0, newInputBuffer.Length);
            _inputBuffer = newInputBuffer;
            processInputBuffer();
        }

        //*********************************************************
        // Networking / Send data
        //*********************************************************

        void sendToServer(OzPDU_OZX pdu)
        {
            var bytes = pdu.Encode();
            lock (_sendLock)
            {
                beginSend(bytes);
            }
        }

        void beginSend(byte[] bytes)
        {
            try
            {
                var ns = client.GetStream();
                ns.BeginWrite(bytes, 0, bytes.Length, endSend, bytes);
            }
            catch (Exception e)
            {
                disconnect("Send failed.");
            }
        }

        void endSend(IAsyncResult result)
        {
            var bytes = (byte[])result.AsyncState;
        }

        //*********************************************************
        // Protocol / Send login
        //*********************************************************

        void sendlogin()
        {
            var pdu = new OzPDU_OZX_Login();
            pdu.Username = Username;
            pdu.Password = Password;
            sendToServer(pdu);
        }

        //*********************************************************
        // Protocol / Send message
        //*********************************************************

        public void Send(OzxMessage message)
        {
            if (client == null || _connectionInProgress)
            {
                OnMessageNotAcceptedForDelivery(this, new OzxArgs<string, string>(message.ID, "Client not connected."));
                return;
            }

            try
            {
                var pdu = new OzPDU_OZX_Send();
                pdu.Message = message;
                sendToServer(pdu);
            }
            catch (Exception ex)
            {
                OnMessageNotAcceptedForDelivery(this, new OzxArgs<string, string>(message.ID, "Client not connected. " + ex.Message));
            }
        }

        //*********************************************************
        // Protocol / receive
        //*********************************************************

        void onPduReceived(OzPDU_OZX pdu)
        {
            switch (pdu.GetCommandID())
            {
                case OzPDU_OZX.Login:
                    //caseCommand_Login((OzPDU_OZX_Login)pdu);
                    break;
                case OzPDU_OZX.Login_Resp:
                    caseCommand_Login_Resp((OzPDU_OZX_Login_Resp)pdu);
                    break;
                case OzPDU_OZX.Keepalive:
                    caseCommand_Keepalive((OzPDU_OZX_Keepalive)pdu);
                    break;
                case OzPDU_OZX.Keepalive_Resp:
                    //caseCommand_Keepalive_Resp((OzPDU_OZX_Keepalive_Resp)pdu);
                    break;
                case OzPDU_OZX.Send:
                    caseCommand_Send((OzPDU_OZX_Send)pdu);
                    break;
                case OzPDU_OZX.Send_Resp:
                    caseCommand_Send_Resp((OzPDU_OZX_Send_Resp)pdu);
                    break;
                case OzPDU_OZX.Send_Success:
                    caseCommand_Send_SubmitSuccess((OzPDU_OZX_Send_SubmitSuccess)pdu);
                    break;
                case OzPDU_OZX.Send_Success_Resp:
                    //caseCommand_Send_Success_Resp((OzPDU_OZX_Send_Success_Resp)pdu);
                    break;
                case OzPDU_OZX.Send_Failed:
                    caseCommand_Send_SubmitFailed((OzPDU_OZX_Send_SubmitFailed)pdu);
                    break;
                case OzPDU_OZX.Send_Failed_Resp:
                    //caseCommand_Send_Failed_Resp((OzPDU_OZX_Send_Failed_Resp)pdu);
                    break;
                case OzPDU_OZX.Send_DeliverySuccess:
                    caseCommand_Send_DeliverySuccess((OzPDU_OZX_Send_DeliverySuccess)pdu);
                    break;
                case OzPDU_OZX.Send_DeliverySuccess_Resp:
                    //caseCommand_Send_DeliverySuccess_Resp((OzPDU_OZX_Send_DeliverySuccess_Resp)pdu);
                    break;
                case OzPDU_OZX.Send_DeliveryFailed:
                    caseCommand_Send_DeliveryFailed((OzPDU_OZX_Send_DeliveryFailed)pdu);
                    break;
                case OzPDU_OZX.Send_DeliveryFailed_Resp:
                    //caseCommand_Send_DeliveryFailed_Resp((OzPDU_OZX_Send_DeliveryFailed_Resp)pdu);
                    break;
                case OzPDU_OZX.Send_Viewed:
                    caseCommand_Send_Viewed((OzPDU_OZX_Send_Viewed)pdu);
                    break;
                case OzPDU_OZX.Send_Viewed_Resp:
                    //caseCommand_Send_Viewed_Resp((OzPDU_OZX_Send_Viewed_Resp)pdu);
                    break;
            }
        }

        //*********************************************************
        // Protocol / Data processing
        //*********************************************************

        void caseCommand_Login_Resp(OzPDU_OZX_Login_Resp pdu)
        {
            if (pdu.Success)
                OnConnected(this, EventArgs.Empty);
            else
                disconnect("Login failed. " + pdu.Message);
        }

        void caseCommand_Keepalive(OzPDU_OZX_Keepalive pdu)
        {
            var pduresp = new OzPDU_OZX_Keepalive_Resp();
            sendToServer(pduresp);
        }

        void caseCommand_Send(OzPDU_OZX_Send pdu)
        {
            OnMessageReceived(this, new OzxArgs<IOzxMessage>(pdu.Message));
            var pduresp = new OzPDU_OZX_Send_Resp();
            pduresp.MessageID = pdu.Message.ID;
            sendToServer(pduresp);
        }

        void caseCommand_Send_Resp(OzPDU_OZX_Send_Resp pdu)
        {
            if (pdu.Accepted)
                OnMessageAcceptedForDelivery(this, new OzxArgs<string>(pdu.MessageID));
            else
                OnMessageNotAcceptedForDelivery(this, new OzxArgs<string, string>(pdu.MessageID, pdu.RejectReason));
        }

        void caseCommand_Send_SubmitSuccess(OzPDU_OZX_Send_SubmitSuccess pdu)
        {
            OnMessageSubmitSuccess(this, new OzxArgs<string, DateTime>(pdu.MessageID, pdu.SendDate));

            var pduresp = new OzPDU_OZX_Send_SubmitSuccess_Resp();
            pduresp.MessageID = pdu.MessageID;
            sendToServer(pduresp);
        }

        void caseCommand_Send_SubmitFailed(OzPDU_OZX_Send_SubmitFailed pdu)
        {
            OnMessageSubmitFailed(this, new OzxArgs<string, DateTime, string>(pdu.MessageID, pdu.SendFailedDate, pdu.Reason));

            var pduresp = new OzPDU_OZX_Send_SubmitFailed_Resp();
            pduresp.MessageID = pdu.MessageID;
            sendToServer(pduresp);
        }

        void caseCommand_Send_DeliverySuccess(OzPDU_OZX_Send_DeliverySuccess pdu)
        {
            OnMessageDeliverySuccess(this, new OzxArgs<string, DateTime>(pdu.MessageID, pdu.DeliveryDate));

            var pduresp = new OzPDU_OZX_Send_DeliverySuccess_Resp();
            pduresp.MessageID = pdu.MessageID;
            sendToServer(pduresp);
        }

        void caseCommand_Send_DeliveryFailed(OzPDU_OZX_Send_DeliveryFailed pdu)
        {
            OnMessageDeliveryFailed(this, new OzxArgs<string, DateTime, string>(pdu.MessageID, pdu.DeliveryFailedDate, pdu.Reason));

            var pduresp = new OzPDU_OZX_Send_DeliveryFailed_Resp();
            pduresp.MessageID = pdu.MessageID;
            sendToServer(pduresp);
        }

        void caseCommand_Send_Viewed(OzPDU_OZX_Send_Viewed pdu)
        {
            OnMessageViewed(this, new OzxArgs<string, DateTime>(pdu.MessageID, pdu.ViewDate));

            var pduresp = new OzPDU_OZX_Send_Viewed_Resp();
            pduresp.MessageID = pdu.MessageID;
            sendToServer(pduresp);
        }

        //*********************************************************
        // Keepalive
        //*********************************************************

        void startKeepAlvie()
        {
            if (_keepAliveTimer != null)
                stopKeepAlive();
            _keepAliveTimer = new Timer(5000);
            _keepAliveTimer.Elapsed += _keepAliveTimer_Elapsed;
            _keepAliveTimer.Enabled = true;
        }

        void stopKeepAlive()
        {
            if (_keepAliveTimer == null) return;
            _keepAliveTimer.Elapsed -= _keepAliveTimer_Elapsed;
            _keepAliveTimer.Enabled = false;
            _keepAliveTimer = null;
        }

        private void _keepAliveTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            if (_lastDataReceived.AddSeconds(20) > DateTime.Now) return;
            var pduresp = new OzPDU_OZX_Keepalive();
            sendToServer(pduresp);
        }

        //*********************************************************
        // Automatic reconnect
        //*********************************************************

        void startReconnectTimer()
        {
            if (_reconnectTimer != null)
                stopReconnectTimer();
            if (!AutoReconnect)
                return;
            _reconnectTimer = new Timer(15000);
            _reconnectTimer.Elapsed += _reconnectTimer_Elapsed;
            _reconnectTimer.Enabled = true;
        }

        void stopReconnectTimer()
        {
            if (_reconnectTimer == null) return;
            _reconnectTimer.Elapsed -= _reconnectTimer_Elapsed;
            _reconnectTimer.Enabled = false;
            _reconnectTimer = null;
        }

        private void _reconnectTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            if (_lastDataReceived.AddSeconds(20) > DateTime.Now) return;
            Connect();
        }
    }
}

